#include "common.h"
#include "shi.h"
#include "parser.h"
#include "util.h"

#include <dlfcn.h>

#include <boost/spirit/iterator/position_iterator.hpp>

#include <xbyak/xbyak.h>

#define NODE_ID(node) ((node).value.id().to_long())
#define NODE_STR(node) string(node.value.begin(), node.value.end())

typedef unsigned char uchar;
typedef unsigned int uint;
typedef tr1::unordered_map<string, vector<int>*> RelocMap;

void* (*shi_dlsym)(void* handle, const char* sym) = &dlsym;

RelocMap::const_iterator add_reloc(const string& s, int a, RelocMap* reloc) {
    RelocMap::iterator iter = reloc->find(s);
    RelocMap::const_iterator r = iter;
    vector<int>* vec;
    if (iter == reloc->end()) {
        vec = new vector<int>;
        r = reloc->insert(make_pair(s, vec)).first;
        //r = reloc->insert(iter, make_pair(s, vec));
    }
    else {
        vec = iter->second;
    }
    vec->push_back(a);
    return r;
}

/* @@@
void error_node(const string& msg, const node& node) {
    const file_position& pos = node.value.begin().get_position();
}
*/

class ShiGen : public Xbyak::CodeGenerator {
private:
    typedef tr1::unordered_map<string, int> SymTable;

public:
    ShiGen(node& program, RelocMap* const_strings0, RelocMap* reloc0 = 0)
        : const_strings(const_strings0), reloc(reloc0) {
        label_id = 0;
        st = 10;
        push(ebp);
        mov(ebp, esp);
        sub(esp, 1000);
        try {
            emit_expr(program);
        }
        catch (Xbyak::Error e) {
            cout << e << endl;
        }
    }

    Xbyak::Address staddr(int index) {
        return dword[ebp-index*4];
    }

    string gen_label() {
        return (format("L%d") % label_id++).str();
    }

    void emit_funcall(const node& expr) {
        string id(NODE_STR(expr.children[0]));
        void* fp = shi_dlsym(RTLD_DEFAULT, id.c_str());
        if (!fp) {
            error("undefined function: " + id);
        }

        bool has_args =
            expr.children.size() != 2 ||
            NODE_ID(expr.children[1]) != Grammar::othersID;

        if (has_args) {
            for (int i = expr.children.size()-1; i > 0; i--) {
                emit_expr(expr.children[i]);
                push(eax);
            }
        }

        if (reloc) {
            add_reloc(id, getSize() + 1, reloc);
        }

        call((int)fp);

        if (has_args) {
            for (int i = expr.children.size()-1; i > 0; i--) {
                pop(edx);
            }
        }
    }

    void emit_pm_op(const string& s, Xbyak::Address dst) {
        assert(s.size() == 1);
        switch (s[0]) {
        case '+':
            add(dst, eax);
            break;
        case '-':
            sub(dst, eax);
            break;
        default:
            error("unknown pm_op");
        }
    }

    void emit_md_op(const string& s, Xbyak::Address dst) {
        assert(s.size() == 1);
        switch (s[0]) {
        case '*':
            mov(edx, dst);
            imul(edx);
            mov(dst, eax);
            break;
        case '/':
            mov(edx, 0);
            mov(ecx, eax);
            mov(eax, dst);
            idiv(ecx);
            mov(dst, eax);
            break;
        case '%':
            mov(edx, 0);
            mov(ecx, eax);
            mov(eax, dst);
            idiv(ecx);
            mov(dst, edx);
            break;
        default:
            error("unknown md_op");
        }
    }

    void emit_cmp_op(const string& s, Xbyak::Address dst) {
        cmp(dst, eax);
        switch (s[0]) {
        case '<':
            if (s[1]) {
                assert(s[1] == '=');
                assert(s.size() == 2);
                setle(dst);
            }
            else {
                assert(s.size() == 1);
                setl(dst);
            }
            break;
        case '>':
            if (s[1]) {
                assert(s[1] == '=');
                assert(s.size() == 2);
                setge(dst);
            }
            else {
                assert(s.size() == 1);
                setg(dst);
            }
            break;
        case '!':
            assert(s[1] == '=');
            assert(s.size() == 2);
            setne(dst);
            break;
        case '=':
            assert(s[1] == '=');
            assert(s.size() == 2);
            sete(dst);
            break;
        default:
            error("unknown cmp_op");
        }
        and(dst, 1);
    }

    void accumulate_binops(void (ShiGen::*emit)(const string&, Xbyak::Address),
                           const vector<node>& nodes) {
        Xbyak::Address tmp = staddr(st++);
        emit_expr(nodes[0]);
        mov(tmp, eax);
        for (size_t i = 1; i < nodes.size(); i+=2) {
            emit_expr(nodes[i+1]);
            (this->*emit)(NODE_STR(nodes[i]), tmp);
        }
        mov(eax, tmp);
        st--;
    }

    void emit_expr(const node& expr) {
        switch (NODE_ID(expr)) {
        case Grammar::int_exprID: {
            mov(eax, lexical_cast<int>(NODE_STR(expr)));
            break;
        }
        case Grammar::str_exprID: {
            string str = NODE_STR(expr);
            replace_all(str, "\\n", "\n");
            RelocMap::const_iterator iter =
                add_reloc(str, getSize() + 1, const_strings);
            mov(eax, (int)iter->first.c_str());
            break;
        }
        case Grammar::identID: {
            string id(NODE_STR(expr));
            SymTable::const_iterator iter = scope.find(id);
            if (iter == scope.end()) {
                error("undefined symbol: " + id);
            }
            mov(eax, staddr(iter->second));
            break;
        }
        case Grammar::assignID: {
            const node& id_node = expr.children[0];
            pair<SymTable::const_iterator, bool> p =
                scope.insert(make_pair(NODE_STR(id_node), st));
            int addr = st;
            if (p.second) {
                st++;
            }
            else {
                addr = p.first->second;
            }
            emit_expr(expr.children[1]);
            mov(staddr(addr), eax);
            break;
        }
        case Grammar::funcallID: {
            emit_funcall(expr);
            break;
        }
        case Grammar::pm_exprID: {
            accumulate_binops(&ShiGen::emit_pm_op, expr.children);
            break;
        }
        case Grammar::md_exprID: {
            accumulate_binops(&ShiGen::emit_md_op, expr.children);
            break;
        }
        case Grammar::cmp_exprID: {
            accumulate_binops(&ShiGen::emit_cmp_op, expr.children);
            break;
        }
        case Grammar::programID:
        case Grammar::scope_stmtID: {
            for_each(expr.children.begin(), expr.children.end(),
                     bind(&ShiGen::emit_expr, this, _1));
            break;
        }
        case Grammar::if_stmtID: {
            emit_expr(expr.children[0]);
            test(eax, eax);
            if (expr.children.size() == 3) {
                string l_else = gen_label();
                string l_fi = gen_label();
                jz(l_else.c_str(), T_NEAR);
                emit_expr(expr.children[1]);
                jmp(l_fi.c_str(), T_NEAR);
                L(l_else.c_str());
                emit_expr(expr.children[2]);
                L(l_fi.c_str());
            }
            else {
                string l_fi = gen_label();
                jz(l_fi.c_str(), T_NEAR);
                emit_expr(expr.children[1]);
                L(l_fi.c_str());
            }
            break;
        }
        case Grammar::while_stmtID: {
            string l_while = gen_label();
            string l_done = gen_label();
            L(l_while.c_str());
            emit_expr(expr.children[0]);
            test(eax, eax);
            jz(l_done.c_str(), T_NEAR);
            emit_expr(expr.children[1]);
            jmp(l_while.c_str());
            L(l_done.c_str());
            break;
        }
        default: {
            error("unknown expr");
        }
        }
    }

    void complete() {
        add(esp, 1000);
        pop(ebp);
        ret();
    }

private:
    SymTable scope;
    int st;
    RelocMap* const_strings;
    int label_id;
    RelocMap* reloc;
};

void compile_program(
    const char* program, char** code_out, int* size_out,
    RelocMap* const_strings_out, RelocMap* reloc_out)
{
    parser_info info = parse_program(program);

    ShiGen* gen = new ShiGen(info.trees[0], const_strings_out, reloc_out);
    gen->complete();
    const uchar* code = gen->getCode();
    int size = gen->getSize();
    {
        FILE* fp = fopen("out", "w");
        fwrite(code, 1, size, fp);
        fclose(fp);
    }

    if (code_out) {
        *code_out = (char*)code;
    }
    if (size_out) {
        *size_out = size;
    }
}

void execute_program(const char* program) {
    char* code;
    int size;
    RelocMap const_strings;
    compile_program(program, &code, &size, &const_strings);

    void (*fp)() = (void (*)())code;
    LOG << endl << "executing..." << endl << endl;
    fp();
}
