#include "shi.h"
#include "util.h"
#include "elf_os.h"

#include <stdlib.h>

#include <tr1/unordered_map>

#include <boost/scoped_ptr.hpp>
#include <boost/foreach.hpp>

#include <elf.h>
#include <sys/stat.h>

using namespace std;
using namespace boost;

/*
  ELF header
  Program header (load)
  Program header (interp)
  Program header (dynamic)
  program
  interp
  padding for align
  Dynamic section
  hash
  dynamic symbol table
  reloc info
  dynamic strtab
 */

template<class C>
class scoped : public C {
public:
    ~scoped() {
        for (typename C::iterator iter = C::begin();
             iter != C::end();
             ++iter)
        {
            delete *iter;
        }
    }
};

typedef tr1::unordered_map<string, vector<int>*> RelocMap;

struct Strtab {
    Strtab() : size(0) {
    }

    int add(const string& s) {
        int r = size;
        strs.push_back(s);
        size += s.size() + 1;
        return r;
    }

    char* data() {
        char* r = (char*)malloc(size);
        char* p = r;
        BOOST_FOREACH(string s, strs) {
            memcpy(p, s.data(), s.size());
            p += s.size();
            *p = '\0';
            p++;
        }
        return r;
    }

    vector<string> strs;
    int size;
};

class Binary {
public:
    explicit Binary(const string& name)
        : filename(name), total(0) {
        fp = fopen(name.c_str(), "wb");
        if (fp == NULL) {
            error("cannot open " + name);
        }
    }

    ~Binary() {
        fclose(fp);
        chmod(filename.c_str(), 0755);
    }

    template <typename T>
    void output(const T& v) {
        fwrite(&v, sizeof(T), 1, fp);
        total += sizeof(T);
    }

    void output(const string& s) {
        output(s.data(), s.size());
        output("\0", 1);
    }

    void output(const void* buf, int size) {
        fwrite(buf, 1, size, fp);
        total += size;
    }

    void output(int data) {
        fwrite(&data, 4, 1, fp);
        total += 4;
    }

    void align(int a) {
        while (total & (a-1)) {
            output("\0", 1);
        }
    }

private:
    FILE* fp;
    string filename;
    int total;
};

void fill_ehdr(Elf32_Ehdr* ehdr, char abi) {
    memset(ehdr->e_ident, 0, 16);
    ehdr->e_ident[EI_MAG0] = ELFMAG0;
    ehdr->e_ident[EI_MAG1] = ELFMAG1;
    ehdr->e_ident[EI_MAG2] = ELFMAG2;
    ehdr->e_ident[EI_MAG3] = ELFMAG3;
    ehdr->e_ident[EI_CLASS] = ELFCLASS32;
    ehdr->e_ident[EI_DATA] = ELFDATA2LSB;
    ehdr->e_ident[EI_VERSION] = 1;
    ehdr->e_ident[EI_OSABI] = abi;
    ehdr->e_type = 2;
    ehdr->e_machine = 3;
    ehdr->e_version = 1;
    ehdr->e_entry = sizeof(Elf32_Ehdr) + sizeof(Elf32_Phdr) * 3;
    ehdr->e_phoff = sizeof(Elf32_Ehdr);
    ehdr->e_shoff = 0;  // We have no shdr.
    ehdr->e_flags = 0;
    ehdr->e_ehsize = sizeof(Elf32_Ehdr);
    ehdr->e_phentsize = sizeof(Elf32_Phdr);
    ehdr->e_phnum = 3;
    ehdr->e_shentsize = 40;
    ehdr->e_shnum = 0;
    ehdr->e_shstrndx = 0;
}

void fill_phdr(int offset, int size, Elf32_Phdr* phdr) {
    phdr->p_offset = offset;
    phdr->p_vaddr = offset;
    phdr->p_paddr = offset;
    phdr->p_filesz = size;
    phdr->p_memsz = size;
}

Elf32_Dyn* make_dyn(Elf32_Sword tag, Elf32_Word val) {
    Elf32_Dyn* dyn = new Elf32_Dyn;
    dyn->d_tag = tag;
    dyn->d_un.d_val = val;
    return dyn;
}

Elf32_Sym* make_undef_sym(Elf32_Word name) {
    Elf32_Sym* sym = new Elf32_Sym;
    sym->st_name = name;
    sym->st_value = 0;
    sym->st_size = 4;
    sym->st_info = ELF32_ST_INFO(STB_GLOBAL, STT_FUNC);
    sym->st_other = ELF32_ST_VISIBILITY(STV_DEFAULT);
    sym->st_shndx = 0;
    return sym;
}

Elf32_Rel* make_ref(Elf32_Addr offset, Elf32_Word info) {
    Elf32_Rel* rel = new Elf32_Rel;
    rel->r_offset = offset;
    rel->r_info = info;
    return rel;
}

int main(int argc, char* argv[]) {
    if (argc < 2) {
        printf("usage: %s <file> [out]", argv[0]);
        return 1;
    }
    string buf = read_file_with_preprocess(argv[1]);
    // ad-hoc
    buf += "\nexit(0)\n";

    string out_name;
    if (argc > 2) {
        out_name = argv[2];
    }
    else {
        out_name = argv[1];
        size_t dot = out_name.rfind('.');
        if (dot == string::npos) {
            error("target doesn't contain dot.");
        }
        out_name = string(argv[1], argv[1] + dot);
    }

    scoped_ptr<ELFOSInfo> os_info(ELFOSInfo::get(OS));

    char* code;
    int code_size;
    RelocMap const_strings;
    RelocMap reloc;
    compile_program(buf.c_str(), &code, &code_size, &const_strings, &reloc);

    int offset = 0;

    Elf32_Ehdr ehdr;
    fill_ehdr(&ehdr, os_info->os_abi());
    offset += sizeof(Elf32_Ehdr);

    Elf32_Phdr phdr_load;
    phdr_load.p_type = PT_LOAD;
    phdr_load.p_flags = PF_X | PF_W | PF_R;
    phdr_load.p_align = 0;
    Elf32_Phdr phdr_interp;
    phdr_interp.p_type = PT_INTERP;
    phdr_interp.p_flags = PF_R;
    phdr_interp.p_align = 1;
    Elf32_Phdr phdr_dynamic;
    phdr_dynamic.p_type = PT_DYNAMIC;
    phdr_dynamic.p_flags = PF_W | PF_R;
    phdr_dynamic.p_align = 4;
    offset += sizeof(Elf32_Phdr) * 3;

    Strtab strtab;
    strtab.add("");

    scoped<vector<Elf32_Sym*> > syms;
    scoped<vector<Elf32_Rel*> > rels;
    for (RelocMap::iterator iter = reloc.begin();
         iter != reloc.end();
         ++iter)
    {
        const string& name = iter->first;
        vector<int>* addrs = iter->second;
        size_t info = ELF32_R_INFO(syms.size(), R_386_PC32);
        syms.push_back(make_undef_sym(strtab.add(name)));
        BOOST_FOREACH(int addr, *addrs) {
            rels.push_back(make_ref(offset + addr, info));

            int* target = (int*)(code + addr);
            *target = -5;
        }
        delete addrs;
    }
    offset += code_size;

    for (RelocMap::iterator iter = const_strings.begin();
         iter != const_strings.end();
         ++iter)
    {
        const string& name = iter->first;
        vector<int>* addrs = iter->second;
        BOOST_FOREACH(int addr, *addrs) {
            int* target = (int*)(code + addr);
            *target = offset;
        }
        offset += name.size() + 1;
        delete addrs;
    }

    const string& interp = os_info->interp();
    fill_phdr(offset, interp.size() + 1, &phdr_interp);
    offset += interp.size() + 1;

    // align
    offset = ((offset-1) & ~3) + 4;

    scoped<vector<Elf32_Dyn*> > dyns;
    dyns.push_back(make_dyn(DT_NEEDED, strtab.add(os_info->libc())));
    if (os_info->libc2()) {
        dyns.push_back(make_dyn(DT_NEEDED, strtab.add(os_info->libc2())));
    }

    int dyn_size = (dyns.size() + 9) * sizeof(Elf32_Dyn);
    fill_phdr(offset, dyn_size, &phdr_dynamic);
    offset += dyn_size;

    dyns.push_back(make_dyn(DT_HASH, offset));
    offset += 16;  // We'll create invalid hash.

    dyns.push_back(make_dyn(DT_SYMENT, sizeof(Elf32_Sym)));
    dyns.push_back(make_dyn(DT_SYMTAB, offset));
    offset += syms.size() * sizeof(Elf32_Sym);

    dyns.push_back(make_dyn(DT_RELENT, sizeof(Elf32_Rel)));
    dyns.push_back(make_dyn(DT_RELSZ, rels.size() * sizeof(Elf32_Rel)));
    dyns.push_back(make_dyn(DT_REL, offset));
    offset += rels.size() * sizeof(Elf32_Rel);

    dyns.push_back(make_dyn(DT_STRSZ, strtab.size));
    dyns.push_back(make_dyn(DT_STRTAB, offset));
    offset += strtab.size;

    dyns.push_back(make_dyn(DT_NULL, 0));

    fill_phdr(0, offset, &phdr_load);
    phdr_load.p_memsz = 4096;

    Binary binary(out_name);
    binary.output(ehdr);
    binary.output(phdr_load);
    binary.output(phdr_interp);
    binary.output(phdr_dynamic);
    binary.output(code, code_size);
    for (RelocMap::iterator iter = const_strings.begin();
         iter != const_strings.end();
         ++iter)
    {
        binary.output(iter->first);
    }
    binary.output(interp);
    binary.align(4);
    BOOST_FOREACH(Elf32_Dyn* dyn, dyns) {
        binary.output(*dyn);
    }
    // Dummy hash
    binary.output(1);
    binary.output(1);
    binary.output(0);
    binary.output(0);
    BOOST_FOREACH(Elf32_Sym* sym, syms) {
        binary.output(*sym);
    }
    BOOST_FOREACH(Elf32_Rel* rel, rels) {
        binary.output(*rel);
    }
    char* strtab_buf = strtab.data();
    binary.output(strtab_buf, strtab.size);
    free(strtab_buf);
}
